from chalk.features import DataFrame
from chalk.utils.missing_dependency import missing_dependency_exception


def assert_frame_equal(
    left: DataFrame,
    right: DataFrame,
    check_column_order: bool = True,
    check_row_order: bool = True,
):
    """Given two ``DataFrame``s, `left` and `right`, check if `left == right`,
    and raise otherwise.

    Parameters
    ----------
    left
        The dataframe to compare.
    right
        The dataframe to compare with.
    check_column_order
        If False, allows the assert/test to succeed if the required columns are present,
        irrespective of the order in which they appear.
    check_row_order
        If False, allows the assert/test to succeed if the required rows are present,
        irrespective of the order in which they appear; as this requires
        sorting, you cannot set on frames that contain unsortable columns.

    Raises
    ------
    AssertionError
        If `left` does not equal `right`
    """
    try:
        import polars.testing
    except ImportError:
        raise missing_dependency_exception("chalkpy[runtime]")
    return polars.testing.assert_frame_equal(
        left.to_polars().collect(),
        right.to_polars().collect(),
        check_row_order=check_row_order,
        check_column_order=check_column_order,
    )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Mapping, Optional, Union

from chalk.integrations.named import load_integration_variable
from chalk.sql._internal.sql_source import BaseSQLSource, TableIngestMixIn
from chalk.sql.protocols import SQLSourceWithTableIngestProtocol
from chalk.utils.missing_dependency import missing_dependency_exception

if TYPE_CHECKING:
    from sqlalchemy.engine.url import URL


class MySQLSourceImpl(BaseSQLSource, TableIngestMixIn, SQLSourceWithTableIngestProtocol):
    def __init__(
        self,
        host: Optional[str] = None,
        port: Optional[Union[int, str]] = None,
        db: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        name: Optional[str] = None,
    ):
        try:
            import pymysql
        except ModuleNotFoundError:
            raise missing_dependency_exception("chalkpy[mysql]")
        del pymysql
        self.host = host or load_integration_variable(name="MYSQL_HOST", integration_name=name)
        self.port = (
            int(port)
            if port is not None
            else load_integration_variable(name="MYSQL_TCP_PORT", integration_name=name, parser=int)
        )
        self.db = db or load_integration_variable(name="MYSQL_DATABASE", integration_name=name)
        self.user = user or load_integration_variable(name="MYSQL_USER", integration_name=name)
        self.password = password or load_integration_variable(name="MYSQL_PWD", integration_name=name)
        self.ingested_tables: Dict[str, Any] = {}
        BaseSQLSource.__init__(self, name=name)

    def engine_args(self) -> Mapping[str, Any]:
        return dict(
            pool_size=20,
            max_overflow=60,
            pool_pre_ping=True,
        )

    def local_engine_url(self) -> URL:
        from sqlalchemy.engine.url import URL

        return URL.create(
            drivername="mysql+pymysql",
            username=self.user,
            password=self.password,
            host=self.host,
            port=self.port,
            database=self.db,
        )



from os import PathLike
from typing import Optional, Union, overload

from chalk.sql._internal.incremental import IncrementalSettings
from chalk.sql._internal.integrations.bigquery import BigQuerySourceImpl
from chalk.sql._internal.integrations.cloudsql import CloudSQLSourceImpl
from chalk.sql._internal.integrations.mysql import MySQLSourceImpl
from chalk.sql._internal.integrations.postgres import PostgreSQLSourceImpl
from chalk.sql._internal.integrations.redshift import RedshiftSourceImpl
from chalk.sql._internal.integrations.snowflake import SnowflakeSourceImpl
from chalk.sql._internal.integrations.sqlite import SQLiteFileSourceImpl, SQLiteInMemorySourceImpl
from chalk.sql.finalized_query import FinalizedChalkQuery
from chalk.sql.protocols import (
    BaseSQLSourceProtocol,
    ChalkQueryProtocol,
    SQLSourceWithTableIngestProtocol,
    StringChalkQueryProtocol,
    TableIngestProtocol,
)


@overload
def SnowflakeSource() -> BaseSQLSourceProtocol:
    """Connect to the only configured Snowflake database.

    If you have only one Snowflake connection that you'd like
    to add to Chalk, you do not need to specify any arguments
    to construct the source in your code.

    Returns
    -------
    BaseSQLSourceProtocol
        The SQL source for use in Chalk resolvers.

    Examples
    --------
    >>> source = SnowflakeSource()
    """


@overload
def SnowflakeSource(*, name: str) -> BaseSQLSourceProtocol:
    """Chalk's injects environment variables to support data integrations.

    But what happens when you have two data sources of the same kind?
    When you create a new data source from your dashboard,
    you have an option to provide a name for the integration.
    You can then reference this name in the code directly.

    Parameters
    ----------
    name
        Name of the integration, as configured in your dashboard.

    Returns
    -------
    BaseSQLSourceProtocol
        The SQL source for use in Chalk resolvers.

    Examples
    --------
    >>> source = SnowflakeSource(name="RISK")
    """
    ...


@overload
def SnowflakeSource(
    *,
    account_identifier: str = ...,
    warehouse: str = ...,
    user: str = ...,
    password: str = ...,
    db: str = ...,
    schema: str = ...,
    role: str = ...,
) -> BaseSQLSourceProtocol:
    """You can also configure the integration directly using environment
    variables on your local machine or from those added through the
    generic environment variable support (https://docs.chalk.ai/docs/env-vars).

    Parameters
    ----------
    account_identifier
        Your Snowflake account identifier.
    warehouse
        Snowflake warehouse to use.
    user
        Username to connect to Snowflake.
    password
        The password to use.
    db
        Database to use.
    schema
        Snowflake schema in the database to use.
    role
        Snowflake role name to use.

    Returns
    -------
    BaseSQLSourceProtocol
        The SQL source for use in Chalk resolvers.

    Examples
    --------
    >>> import os
    >>> snowflake = SnowflakeSource(
    ...     db=os.getenv("SNOWSQL_DATABASE"),
    ...     schema=os.getenv("SNOWSQL_SCHEMA"),
    ...     role=os.getenv("SNOWSQL_ROLE"),
    ...     warehouse=os.getenv("SNOWSQL_WAREHOUSE"),
    ...     user=os.getenv("SNOWSQL_USER"),
    ...     password=os.getenv("SNOWSQL_PWD"),
    ...     account_identifier=os.getenv("SNOWSQL_ACCOUNT_IDENTIFIER")
    ... )
    """
    ...


def SnowflakeSource(
    *,
    name: Optional[str] = None,
    account_identifier: Optional[str] = None,
    warehouse: Optional[str] = None,
    user: Optional[str] = None,
    password: Optional[str] = None,
    db: Optional[str] = None,
    schema: Optional[str] = None,
    role: Optional[str] = None,
) -> BaseSQLSourceProtocol:
    return SnowflakeSourceImpl(
        name=name,
        account_identifier=account_identifier,
        warehouse=warehouse,
        user=user,
        password=password,
        db=db,
        schema=schema,
        role=role,
    )


@overload
def PostgreSQLSource() -> SQLSourceWithTableIngestProtocol:
    """Connect to the only configured PostgreSQL database.

    If you have only one PostgreSQL connection that you'd like
    to add to Chalk, you do not need to specify any arguments
    to construct the source in your code.

    Returns
    -------
    SQLSourceWithTableIngestProtocol
        The SQL source for use in Chalk resolvers.

    Examples
    --------
    >>> pg = PostgreSQLSource()
    """
    ...


@overload
def PostgreSQLSource(
    *,
    host: str = ...,
    port: Union[int, str] = ...,
    db: str = ...,
    user: str = ...,
    password: str = ...,
) -> SQLSourceWithTableIngestProtocol:
    """You can also configure the integration directly using environment
    variables on your local machine or from those added through the
    generic environment variable support (https://docs.chalk.ai/docs/env-vars).

    Parameters
    ----------
    host
        Name of host to connect to.
    port
        The port number to connect to at the server host.
    db
        The database name.
    user
        PostgreSQL username to connect as.
    password
        The password to be used if the server demands password authentication.

    Returns
    -------
    SQLSourceWithTableIngestProtocol
        The SQL source for use in Chalk resolvers.

    Examples
    --------
    >>> import os
    >>> pg = PostgreSQLSource(
    ...     host=os.getenv("PGHOST"),
    ...     port=os.getenv("PGPORT"),
    ...     db=os.getenv("PGDATABASE"),
    ...     user=os.getenv("PGUSER"),
    ...     password=os.getenv("PGPASSWORD"),
    ... )

    >>> from chalk.features import online
    >>> @online
    ... def resolver_fn() -> User.name:
    ...     return pg.query_string("select name from users where id = 4").one()
    """
    ...


@overload
def PostgreSQLSource(*, name: str) -> SQLSourceWithTableIngestProtocol:
    """If you have only one PostgreSQL integration, there's no need to provide
    a distinguishing name.

    But what happens when you have two data sources of the same kind?
    When you create a new data source from your dashboard,
    you have an option to provide a name for the integration.
    You can then reference this name in the code directly.

    Parameters
    ----------
    name
        Name of the integration, as configured in your dashboard.

    Returns
    -------
    SQLSourceWithTableIngestProtocol
        The SQL source for use in Chalk resolvers.

    Examples
    --------
    >>> source = PostgreSQLSource(name="RISK")
    """
    ...


def PostgreSQLSource(
    *,
    host: Optional[str] = None,
    port: Optional[Union[int, str]] = None,
    db: Optional[str] = None,
    user: Optional[str] = None,
    password: Optional[str] = None,
    name: Optional[str] = None,
) -> TableIngestProtocol:
    """You can also configure the integration directly using environment
    variables on your local machine or from those added through the
    generic environment variable support (https://docs.chalk.ai/docs/env-vars).

    Parameters
    ----------
    host
        Name of host to connect to.
    port
        The port number to connect to at the server host.
    db
        The database name.
    user
        PostgreSQL username to connect as.
    password
        The password to be used if the server demands password authentication.
    name
        The name of the integration, as configured in your dashboard.

    Returns
    -------
    SQLSourceWithTableIngestProtocol
        The SQL source for use in Chalk resolvers.
    """
    return PostgreSQLSourceImpl(host, port, db, user, password, name)


@overload
def MySQLSource(
    *,
    host: str = ...,
    port: Union[int, str] = ...,
    db: str = ...,
    user: str = ...,
    password: str = ...,
) -> SQLSourceWithTableIngestProtocol:
    """

    Parameters
    ----------
    host
    port
    db
    user
    password

    Returns
    -------

    """
    ...


@overload
def MySQLSource() -> SQLSourceWithTableIngestProtocol:
    """If you have only one MySQL connection that you'd like
    to add to Chalk, you do not need to specify any arguments
    to construct the source in your code.

    Returns
    -------
    SQLSourceWithTableIngestProtocol
        The SQL source for use in Chalk resolvers.

    Examples
    --------
    >>> mysql = MySQLSource()
    """
    ...


@overload
def MySQLSource(*, name: str) -> SQLSourceWithTableIngestProtocol:
    """If you have only one MySQL integration, there's no need to provide
    a distinguishing name.

    But what happens when you have two data sources of the same kind?
    When you create a new data source from your dashboard,
    you have an option to provide a name for the integration.
    You can then reference this name in the code directly.

    Parameters
    ----------
    name
        Name of the integration, as configured in your dashboard.

    Returns
    -------
    SQLSourceWithTableIngestProtocol
        The SQL source for use in Chalk resolvers.

    Examples
    --------
    >>> source = MySQLSource(name="RISK")
    """
    ...


def MySQLSource(
    *,
    host: Optional[str] = None,
    port: Optional[Union[int, str]] = None,
    db: Optional[str] = None,
    user: Optional[str] = None,
    password: Optional[str] = None,
    name: Optional[str] = None,
) -> SQLSourceWithTableIngestProtocol:
    return MySQLSourceImpl(host, port, db, user, password, name)


def SQLiteInMemorySource(name: Optional[str] = None) -> SQLSourceWithTableIngestProtocol:
    return SQLiteInMemorySourceImpl(name)


def SQLiteFileSource(filename: Union[str, PathLike], name: Optional[str] = None) -> SQLSourceWithTableIngestProtocol:
    return SQLiteFileSourceImpl(filename, name)


@overload
def RedshiftSource(
    *,
    host: str = ...,
    db: str = ...,
    user: str = ...,
    password: str = ...,
) -> BaseSQLSourceProtocol:
    ...


@overload
def RedshiftSource(*, name: str) -> BaseSQLSourceProtocol:
    """If you have only one Redshift integration, there's no need to provide
    a distinguishing name.

    But what happens when you have two data sources of the same kind?
    When you create a new data source from your dashboard,
    you have an option to provide a name for the integration.
    You can then reference this name in the code directly.

    Parameters
    ----------
    name
        Name of the integration, as configured in your dashboard.

    Returns
    -------
    BaseSQLSourceProtocol
        The SQL source for use in Chalk resolvers.

    Examples
    --------
    >>> source = RedshiftSource(name="RISK")
    """
    ...


@overload
def RedshiftSource() -> BaseSQLSourceProtocol:
    """If you have only one Redshift connection that you'd like
    to add to Chalk, you do not need to specify any arguments
    to construct the source in your code.

    Returns
    -------
    BaseSQLSourceProtocol
        The SQL source for use in Chalk resolvers.

    Examples
    --------
    >>> source = RedshiftSource()
    """
    ...


def RedshiftSource(
    *,
    host: Optional[str] = None,
    db: Optional[str] = None,
    user: Optional[str] = None,
    password: Optional[str] = None,
    name: Optional[str] = None,
) -> BaseSQLSourceProtocol:
    return RedshiftSourceImpl(host, db, user, password, name)


@overload
def BigQuerySource() -> BaseSQLSourceProtocol:
    """If you have only one BigQuery connection that you'd like
    to add to Chalk, you do not need to specify any arguments
    to construct the source in your code.

    Returns
    -------
    BaseSQLSourceProtocol
        The SQL source for use in Chalk resolvers.

    Examples
    --------
    >>> source = BigQuerySource()
    """
    ...


@overload
def BigQuerySource(*, name: str) -> BaseSQLSourceProtocol:
    """If you have only one BigQuery integration, there's no need to provide
    a distinguishing name.

    But what happens when you have two data sources of the same kind?
    When you create a new data source from your dashboard,
    you have an option to provide a name for the integration.
    You can then reference this name in the code directly.

    Parameters
    ----------
    name
        Name of the integration, as configured in your dashboard.

    Returns
    -------
    BaseSQLSourceProtocol
        The SQL source for use in Chalk resolvers.

    Examples
    --------
    >>> source = BigQuerySource(name="RISK")
    """
    ...


def BigQuerySource(
    *,
    name: Optional[str] = None,
    project: Optional[str] = None,
    dataset: Optional[str] = None,
    location: Optional[str] = None,
    credentials_base64: Optional[str] = None,
    credentials_path: Optional[str] = None,
) -> BaseSQLSourceProtocol:
    return BigQuerySourceImpl(
        name=name,
        project=project,
        dataset=dataset,
        location=location,
        credentials_base64=credentials_base64,
        credentials_path=credentials_path,
    )


@overload
def CloudSQLSource() -> BaseSQLSourceProtocol:
    """If you have only one CloudSQL connection that you'd like
    to add to Chalk, you do not need to specify any arguments
    to construct the source in your code.

    Returns
    -------
    BaseSQLSourceProtocol
        The SQL source for use in Chalk resolvers.

    Examples
    --------
    >>> source = CloudSQLSource()
    """
    ...


@overload
def CloudSQLSource(*, name: str) -> BaseSQLSourceProtocol:
    """If you have only one CloudSQL integration, there's no need to provide
    a distinguishing name.

    But what happens when you have two data sources of the same kind?
    When you create a new data source from your dashboard,
    you have an option to provide a name for the integration.
    You can then reference this name in the code directly.

    Parameters
    ----------
    name
        Name of the integration, as configured in your dashboard.

    Returns
    -------
    BaseSQLSourceProtocol
        The SQL source for use in Chalk resolvers.

    Examples
    --------
    >>> source = CloudSQLSource(name="RISK")
    """
    ...


def CloudSQLSource(
    *,
    name: Optional[str] = None,
    instance_name: Optional[str] = None,
    db: Optional[str] = None,
    user: Optional[str] = None,
    password: Optional[str] = None,
) -> BaseSQLSourceProtocol:
    return CloudSQLSourceImpl(
        name=name,
        instance_name=instance_name,
        db=db,
        user=user,
        password=password,
    )


__all__ = [
    "BaseSQLSourceProtocol",
    "BigQuerySource",
    "CloudSQLSource",
    "MySQLSource",
    "PostgreSQLSource",
    "RedshiftSource",
    "SQLiteFileSource",
    "SQLiteInMemorySource",
    "SnowflakeSource",
    "FinalizedChalkQuery",
    "ChalkQueryProtocol",
    "StringChalkQueryProtocol",
    "TableIngestProtocol",
    "SQLSourceWithTableIngestProtocol",
    "IncrementalSettings",
]



from datetime import datetime
from typing import Any

from typing_extensions import Annotated

from chalk.features.feature_field import Feature
from chalk.features.feature_wrapper import unwrap_feature

FeatureTime = Annotated[datetime, "__chalk_ts__"]
"""Marker for a FeatureTime.

>>> from chalk.features import features
>>> @features
... class User:
...     updated_at: FeatureTime
"""


def feature_time() -> Any:
    """Declare a FeatureTime (deprecated).

    Deprecated in favor of
    >>> from chalk.features import features
    >>> @features
    ... class User:
    ...     updated_at: FeatureTime
    """
    return Feature(typ=datetime, is_feature_time=True)


def is_feature_time(f: Any) -> bool:
    """Determine whether a feature is a feature time.

    Parameters
    ----------
    f
        A feature (i.e. User.ts)

    Returns
    -------
    bool
        True if the feature is a `FeatureTime` and False otherwise
    """
    return unwrap_feature(f).is_feature_time



from __future__ import annotations

import base64
import collections.abc
from datetime import date, datetime, time, timedelta
from typing import Any, List, Optional, Type, cast

import cattrs
import dateutil.parser
import isodate
from typing_extensions import get_args, get_origin, get_type_hints, is_typeddict

from chalk.features._encoding.primitive import TPrimitive
from chalk.utils.json import TJSON

__all__ = ["unstructure_primitive_to_json", "structure_json_to_primitive"]


_json_converter = cattrs.Converter()


def unstructure_primitive_to_json(val: TPrimitive) -> TJSON:
    return _json_converter.unstructure(val)


def structure_json_to_primitive(val: TJSON, typ: Type[TPrimitive]) -> TPrimitive:
    return _json_converter.structure(val, typ)


#######
# Dicts
#######

# All structs get mapped to a typeddict primitive type
# However, on the wire, we serialize typedicts as lists by field order, similar
# to how structs are commonly serialized by values
# Hence, the JSON converter needs to convert all dicts into lists of the values
# In py3.7+, all dicts are ordered. So, simply iterating over the dict values is sufficient
_json_converter.register_unstructure_hook(dict, lambda x: [_json_converter.unstructure(y) for y in x.values()])


def _structure_typeddict(obj: Any, typ: Type):
    if not isinstance(obj, collections.abc.Sequence):
        raise TypeError(f"Expected structs to be serialized as lists. Object `{obj}` is not a sequence.")
    type_hints = get_type_hints(typ)
    if len(type_hints) != len(obj):
        raise TypeError(
            f"Unable to structure object `{obj}` of size {len(obj)} into type `{typ.__name__}` of size {len(type_hints)}. Size mismatch."
        )

    kwargs = {
        field_name: None if x is None else _json_converter.structure(x, type_hints[field_name])
        for (x, field_name) in zip(obj, type_hints.keys())
    }
    return typ(**kwargs)


_json_converter.register_structure_hook_func(is_typeddict, _structure_typeddict)


#######
# Lists
#######

# The default unstructure hook is fine
# However, when structuring, we want to allow None for annotations of list[...]


def _structure_list(obj: Optional[List], typ: Type[List]) -> List:
    args = get_args(typ)

    if len(args) < 1:
        raise TypeError(
            f"{typ} types must be parameterized with the type of the contained value -- for example, `{typ}[int]`"
        )
    if len(args) > 1:
        raise TypeError(f"{typ} should be parameterized with only one type")
    if obj is None:
        return cast(List, None)
    inner_typ = args[0]
    return [_json_converter.structure(x, inner_typ) for x in obj]


def _is_list(typ: Type):
    origin = get_origin(typ)
    return origin in (list, List)


_json_converter.register_structure_hook_func(_is_list, _structure_list)

##########
# Datetime
##########

_json_converter.register_unstructure_hook(datetime, lambda x: x.isoformat())


def _structure_datetime(obj: Any, typ: Type):
    if not isinstance(obj, str):
        raise TypeError(
            f"Datetime values must be serialized as ISO strings. Instead, received value '{obj}' of type `{type(obj).__name__}`"
        )
    return dateutil.parser.parse(obj)


_json_converter.register_structure_hook(datetime, _structure_datetime)


######
# Date
######


def _is_date(x: Type):
    return isinstance(x, type) and issubclass(x, date) and not issubclass(x, datetime)


_json_converter.register_unstructure_hook_func(
    _is_date,
    lambda x: x.isoformat(),
)


def _structure_date(obj: Any, typ: Type):
    if not isinstance(obj, str):
        raise TypeError(
            f"Date values must be serialized as ISO strings. Instead, received value '{obj}' of type `{type(obj).__name__}`"
        )
    return datetime.combine(isodate.parse_date(obj), time())


_json_converter.register_structure_hook_func(
    _is_date,
    _structure_date,
)


######
# Time
######

_json_converter.register_unstructure_hook(time, lambda x: x.isoformat())


def _structure_time(obj: Any, typ: Type):
    if not isinstance(obj, str):
        raise TypeError(
            f"Time values must be serialized as ISO strings. Instead, received value '{obj}' of type `{type(obj).__name__}`"
        )
    return isodate.parse_time(obj)


_json_converter.register_structure_hook(time, _structure_time)

########
# Binary
########

_json_converter.register_unstructure_hook(bytes, lambda x: base64.b64encode(x).decode("utf8"))


def _structure_bytes(obj: Any, typ: Type):
    if isinstance(obj, str):
        return base64.b64decode(obj)
    if isinstance(obj, bytes):
        return obj
    raise TypeError(
        f"Byte values must be bytes objects or Base64-encoded strings. Instead, received value '{obj}' of type `{type(obj).__name__}`"
    )


_json_converter.register_structure_hook(bytes, _structure_bytes)

_json_converter.register_unstructure_hook(timedelta, isodate.duration_isoformat)

###########
# Timedelta
###########


def _structure_timedelta(obj: Any, typ: Type):
    if not isinstance(obj, str):
        raise TypeError(
            f"Timedelta values should be serialized as strings. Instead, received value '{obj}' of type `{type(obj).__name__}`"
        )
    return isodate.parse_duration(obj)


_json_converter.register_structure_hook(timedelta, _structure_timedelta)

####################
# Int/float/str/bool
####################


def _structure_basic(obj: Any, typ: Type):
    if obj is None:
        # Always allow None, even if the field is non-optional
        return None
    if issubclass(typ, bool):
        # For booleans, we are more strict than just doing a simple cast, since
        # bool("random string") or bool(100) should raise an exception, not be True
        if obj in (1, True):
            return True
        if obj in (0, False):
            return False
        raise TypeError(f"Cannot convert '{obj}' to a Boolean. Valid values are 1, True, 0, or False.")
    return typ(obj)


_json_converter.register_structure_hook(int, _structure_basic)
_json_converter.register_structure_hook(float, _structure_basic)
_json_converter.register_structure_hook(str, _structure_basic)
_json_converter.register_structure_hook(bool, _structure_basic)



import inspect
from typing import Any, Callable, List, Literal, Optional, Type, TypeVar, Union

from pydantic import BaseModel
from typing_extensions import ParamSpec

from chalk.state import KeyedState
from chalk.streams._file_source import FileSource
from chalk.streams._kafka_source import KafkaSource
from chalk.streams._windows import Windowed, get_duration_secs, get_name_with_duration, windowed
from chalk.streams.base import StreamSource
from chalk.utils import MachineType

__all__ = [
    "FileSource",
    "KafkaSource",
    "KeyedState",
    "StreamSource",
    "Windowed",
    "stream",
    "windowed",
    "get_name_with_duration",
    "get_duration_secs",
]

P = ParamSpec("P")
T = TypeVar("T")

MessageType = TypeVar("MessageType", bound=BaseModel)


def stream(
    *,
    source: StreamSource,
    mode: Optional[Literal["continuous", "tumbling"]] = None,
    environment: Optional[Union[List[str], str]] = None,
    machine_type: Optional[MachineType] = None,
    message: Optional[Type[Any]] = None,
    owner: Optional[str] = None,
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    """Decorator to create a stream resolver.

    Parameters
    ----------
    source
        The streaming source, e.g. `KafkaSource()`
    mode
        The streaming mode, either "continuous" or "tumbling".
    environment
        Environments are used to trigger behavior in different deployments
        such as staging, production, and local development.

        Environment can take one of three types:
            - None (default) - candidate to run in every environment
            - str - run only in this environment
            - list[str] - run in any of the specified environment and no others

        Read more at https://docs.chalk.ai/docs/resolver-environments
    machine_type
        You can optionally specify that resolvers need to run
        on a machine other than the default. Must be configured
        in your deployment.
    message
        A `dataclass` describing the JSON format of the messages on
        the topic.
    owner
        Allows you to specify an individual or team who is responsible
        for this resolver. The Chalk Dashboard will display this field,
        and alerts can be routed to owners.

    Returns
    -------
    Callable[[Any, ...], Any]
        A callable function! You can unit-test resolvers as you would
        unit-test any other code.

        Read more at https://docs.chalk.ai/docs/unit-tests
    """
    caller_frame = inspect.stack()[1]
    caller_filename = caller_frame.filename
    caller_globals = caller_frame.frame.f_globals
    caller_locals = caller_frame.frame.f_locals
    from chalk.features.resolver import parse_and_register_stream_resolver

    def decorator(fn: Callable[P, T]) -> Callable[P, T]:
        return parse_and_register_stream_resolver(
            caller_globals=caller_globals,
            caller_locals=caller_locals,
            fn=fn,
            source=source,
            mode=mode,
            caller_filename=caller_filename,
            environment=environment,
            machine_type=machine_type,
            message=message,
            owner=owner,
        )

    return decorator



import collections.abc
from typing import Any, Dict, Iterable, List, Optional, Sequence, Set, Tuple, TypeVar, Union, cast, overload

from typing_extensions import get_args, get_origin

try:
    from types import UnionType
except ImportError:
    UnionType = Union

_NoneType = type(None)

T = TypeVar("T")


def unwrap_optional_if_needed(typ: Any):
    origin = get_origin(typ)
    args = get_args(typ)
    if origin in (Union, UnionType) and len(args) == 2 and any(d == _NoneType for d in args):
        return next(m for m in get_args(typ) if m != _NoneType)

    return typ


def is_namedtuple(value: Any) -> bool:
    """Infer whether value is a NamedTuple."""
    # From https://github.com/pola-rs/polars/blob/5f3e332fb2a653064f083b02949c527e0ec0afda/py-polars/polars/internals/construction.py#L78
    return all(hasattr(value, attr) for attr in ("_fields", "_field_defaults", "_replace"))


def flatten(v: Sequence[Union[T, Sequence[Union[T, Sequence[T]]]]]) -> List[T]:
    ret = []
    for x in v:
        if isinstance(x, collections.abc.Sequence) and not isinstance(x, (str, bytes, bytearray)):
            ret.extend(flatten(x))
        else:
            ret.append(x)
    return ret


@overload
def chunks(lst: List[T], n: int) -> Iterable[List[T]]:
    ...


@overload
def chunks(lst: Set[T], n: int) -> Iterable[Set[T]]:
    ...


@overload
def chunks(lst: Tuple[T, ...], n: int) -> Iterable[Tuple[T, ...]]:
    ...


@overload
def chunks(lst: Iterable[T], n: int) -> Iterable[Iterable[T]]:
    ...


def chunks(lst: Iterable[T], n: int) -> Iterable[Iterable[T]]:
    """Yield successive n-sized chunks from ``lst``, potentially lazily.

    Chunks are generated up-front if ``lst`` is a list, tuple, or set. In this case,
    the result will be an iterator of n-sized chunks of the underlying collection type.

    Otherwise, the chunks and the contents of each chunk are generated lazily,
    where at most one sample from ``lst`` is read in advance.

    Parameters
    ----------
    lst
        The collection
    n
        The chunk size. If <=0, then everything will be yielded back as one chunk.
    """
    if n <= 0:
        yield lst

    iterator = iter(lst)

    # Greedily loading the next sample to detect if we exhausted the iterable, so we don't yield an empty chunk at the end
    have_sample_to_yield = True
    try:
        next_sample = next(iterator)
    except StopIteration:
        have_sample_to_yield = False
        return

    def _get_chunk() -> Iterable[T]:
        """Get a chunk of size n from the generator, or raises StopIteration of the lst is empty"""
        nonlocal next_sample, have_sample_to_yield
        for _ in range(n):
            yield next_sample

            try:
                next_sample = next(iterator)
            except StopIteration:
                have_sample_to_yield = False
                break

    while have_sample_to_yield:
        chunk = _get_chunk()
        if isinstance(lst, (tuple, set, list)):
            # If the underlying collection is already allocated, then allocate the entire chunk at once
            # For backwards compatibility
            yield type(lst)(chunk)
        else:
            yield chunk


def flatten_sets(v: List[Set[T]]) -> Set[T]:
    return set().union(*v)


def ensure_tuple(x: Union[T, Sequence[T], Dict[Any, T], None]) -> Tuple[T, ...]:
    """Converts ``x`` into a tuple.
    * If ``x`` is ``None``, then ``tuple()`` is returned.
    * If ``x`` is a tuple, then ``x`` is returned as-is.
    * If ``x`` is a list, then ``tuple(x)`` is returned.
    * If ``x`` is a dict, then ``tuple(v for v in x.values())`` is returned.
    Otherwise, a single element tuple of ``(x,)`` is returned.

    Parameters
    ----------
    x
        The input to convert into a tuple.

    Returns
    -------
    tuple
        A tuple of ``x``.
    """
    # From https://github.com/mosaicml/composer/blob/020ca02e3848ee8fb6b7fff0c8123f597b05be8a/composer/utils/iter_helpers.py#L40
    if x is None:
        return ()
    if isinstance(x, (str, bytes, bytearray)):
        return (cast(T, x),)
    if isinstance(x, collections.abc.Sequence):
        return tuple(x)
    if isinstance(x, dict):
        return tuple(x.values())
    return (x,)


def get_unique_item(collection: Iterable[Optional[T]], name: str) -> T:
    item = None
    for x in collection:
        if x is None:
            raise ValueError(f"Item in {name} is None")
        if item is not None:
            if x != item:
                raise ValueError(f"Multiple values in {name} are not permitted. Found {x}, {item}")
        item = x
    if item is None:
        raise ValueError(f"There should be at least one item in {name}")
    return item



from typing import Literal, Union

MissingValueStrategy = Union[
    Literal["error"],  # Raise a TypeError if missing values are found
    # Coerce missing values to the default value for the feature, if one is defined. Otherwise, raise a TypeError
    Literal["default_or_error"],
    # Coerce missing values to the default value for the feature, if one is defined. Otherwise, treat missing values as valid.
    Literal["default_or_allow"],
    Literal["allow"],  # Treat missing values as valid values
]



import dataclasses
import zlib
from copy import deepcopy
from enum import Enum
from typing import Callable, ClassVar, List, Literal, Optional, Set, Tuple, TypeVar, Union, overload

from chalk import Resolver
from chalk.features import FeatureWrapper
from chalk.utils.duration import parse_chalk_duration

T = TypeVar("T")


class MetricKind(str, Enum):
    FEATURE_REQUEST_COUNT = "FEATURE_REQUEST_COUNT"
    FEATURE_LATENCY = "FEATURE_LATENCY"
    FEATURE_STALENESS = "FEATURE_STALENESS"
    FEATURE_VALUE = "FEATURE_VALUE"
    FEATURE_WRITE = "FEATURE_WRITE"
    FEATURE_NULL_RATIO = "FEATURE_NULL_RATIO"

    RESOLVER_REQUEST_COUNT = "RESOLVER_REQUEST_COUNT"
    RESOLVER_LATENCY = "RESOLVER_LATENCY"
    RESOLVER_SUCCESS_RATIO = "RESOLVER_SUCCESS_RATIO"

    QUERY_COUNT = "QUERY_COUNT"
    QUERY_LATENCY = "QUERY_LATENCY"
    QUERY_SUCCESS_RATIO = "QUERY_SUCCESS_RATIO"

    BILLING_INFERENCE = "BILLING_INFERENCE"
    BILLING_CRON = "BILLING_CRON"
    BILLING_MIGRATION = "BILLING_MIGRATION"

    CRON_COUNT = "CRON_COUNT"
    CRON_LATENCY = "CRON_LATENCY"

    STREAM_MESSAGES_PROCESSED = "STREAM_MESSAGES_PROCESSED"
    STREAM_MESSAGE_LATENCY = "STREAM_MESSAGE_LATENCY"

    STREAM_WINDOWS_PROCESSED = "STREAM_WINDOWS_PROCESSED"
    STREAM_WINDOW_LATENCY = "STREAM_WINDOW_LATENCY"


class FilterKind(str, Enum):
    FEATURE_STATUS = "FEATURE_STATUS"
    FEATURE_NAME = "FEATURE_NAME"
    FEATURE_TAG = "FEATURE_TAG"

    RESOLVER_STATUS = "RESOLVER_STATUS"
    RESOLVER_NAME = "RESOLVER_NAME"
    RESOLVER_TAG = "RESOLVER_TAG"

    CRON_STATUS = "CRON_STATUS"
    MIGRATION_STATUS = "MIGRATION_STATUS"

    ONLINE_OFFLINE = "ONLINE_OFFLINE"
    CACHE_HIT = "CACHE_HIT"
    OPERATION_ID = "OPERATION_ID"

    QUERY_NAME = "QUERY_NAME"
    QUERY_STATUS = "QUERY_STATUS"

    IS_NULL = "IS_NULL"


ResolverType = Literal["online", "offline", "stream"]
FeatureNameType = Union[FeatureWrapper, str]
ResolverNameType = Union[Resolver, str]


class ComparatorKind(str, Enum):
    EQ = "EQ"
    NEQ = "NEQ"
    ONE_OF = "ONE_OF"


class WindowFunctionKind(str, Enum):
    COUNT = "COUNT"
    MEAN = "MEAN"
    SUM = "SUM"
    MIN = "MIN"
    MAX = "MAX"

    PERCENTILE_99 = "PERCENTILE_99"
    PERCENTILE_95 = "PERCENTILE_95"
    PERCENTILE_75 = "PERCENTILE_75"
    PERCENTILE_50 = "PERCENTILE_50"
    PERCENTILE_25 = "PERCENTILE_25"
    PERCENTILE_5 = "PERCENTILE_5"

    ALL_PERCENTILES = "ALL_PERCENTILES"

    @classmethod
    def has_member_key(cls, key):
        return key in cls.__members__


class GroupByKind(str, Enum):
    FEATURE_STATUS = "FEATURE_STATUS"
    FEATURE_NAME = "FEATURE_NAME"
    IS_NULL = "IS_NULL"

    RESOLVER_STATUS = "RESOLVER_STATUS"
    RESOLVER_NAME = "RESOLVER_NAME"

    QUERY_STATUS = "QUERY_STATUS"
    QUERY_NAME = "QUERY_NAME"

    ONLINE_OFFLINE = "ONLINE_OFFLINE"
    CACHE_HIT = "CACHE_HIT"


class MetricFormulaKind(str, Enum):
    SUM = "SUM"
    TOTAL_RATIO = "TOTAL_RATIO"
    RATIO = "RATIO"
    PRODUCT = "PRODUCT"
    ABS = "ABS"
    KS_STAT = "KS_STAT"
    KS_TEST = "KS_TEST"
    KS_THRESHOLD = "KS_THRESHOLD"
    TIME_OFFSET = "TIME_OFFSET"


class AlertSeverityKind(str, Enum):
    CRITICAL = "CRITICAL"
    ERROR = "ERROR"
    WARNING = "WARNING"
    INFO = "INFO"


class ThresholdPosition(str, Enum):
    ABOVE = "ABOVE"
    BELOW = "BELOW"


class ChartLinkKind(str, Enum):
    resolver = "resolver"
    feature = "feature"
    query = "query"
    manual = "manual"


@dataclasses.dataclass
class MetricFilter:
    kind: FilterKind
    comparator: ComparatorKind
    value: List[str]


@dataclasses.dataclass
class ThresholdFunction:
    lhs: "Series"
    operation: str
    rhs: float


@dataclasses.dataclass
class SingleSeriesOperand:
    operand: int


@dataclasses.dataclass
class MultiSeriesOperand:
    operands: List[int]


@dataclasses.dataclass
class DatasetFeatureOperand:
    dataset: str
    feature: str


class Series:
    def __init__(
        self,
        name: str,
        metric: Union[MetricKind, str],
        window_function: Optional[Union[WindowFunctionKind, str]] = None,
        time_shift: Optional[int] = None,
    ):
        self._name = name
        self._metric = MetricKind(metric.upper()) if metric else None
        self._filters: List[MetricFilter] = []
        if window_function:
            if WindowFunctionKind.has_member_key(window_function.upper()):
                window_function_kind = WindowFunctionKind(window_function.upper())
            else:
                window_function_kind = self._get_window_function_type(window_function.upper())
                if not window_function_kind:
                    raise ValueError(f"'window_function' value '{window_function}' 'invalid for WindowFunctionKind")
        else:
            window_function_kind = None
        self._window_function = window_function_kind
        self._group_by: List[GroupByKind] = []
        self._time_shift = time_shift
        self._entity_kind = ChartLinkKind.manual
        self._entity_id = None

    @staticmethod
    def _get_window_function_type(key: str) -> Union[WindowFunctionKind, None]:
        window_function_value_dict = {
            "99%": WindowFunctionKind.PERCENTILE_99,
            "95%": WindowFunctionKind.PERCENTILE_95,
            "75%": WindowFunctionKind.PERCENTILE_75,
            "50%": WindowFunctionKind.PERCENTILE_50,
            "25%": WindowFunctionKind.PERCENTILE_25,
            "5%": WindowFunctionKind.PERCENTILE_5,
            "99": WindowFunctionKind.PERCENTILE_99,
            "95": WindowFunctionKind.PERCENTILE_95,
            "75": WindowFunctionKind.PERCENTILE_75,
            "50": WindowFunctionKind.PERCENTILE_50,
            "25": WindowFunctionKind.PERCENTILE_25,
            "5": WindowFunctionKind.PERCENTILE_5,
            "ALL": WindowFunctionKind.ALL_PERCENTILES,
        }
        return window_function_value_dict.get(key)

    @staticmethod
    def _get_filter_string_type(key: str) -> Union[FilterKind, None]:
        string_dict = {
            "feature_tag": FilterKind.FEATURE_TAG,
            "resolver_tag": FilterKind.RESOLVER_TAG,
            "operation_id": FilterKind.OPERATION_ID,
            "query_name": FilterKind.QUERY_NAME,
        }
        return string_dict.get(key)

    @staticmethod
    def _get_filter_status_type(key: str) -> Union[FilterKind, None]:
        status_dict = {
            "feature": FilterKind.FEATURE_STATUS,
            "resolver": FilterKind.RESOLVER_STATUS,
            "cron": FilterKind.CRON_STATUS,
            "migration": FilterKind.MIGRATION_STATUS,
            "query": FilterKind.QUERY_STATUS,
        }
        return status_dict.get(key)

    def where(
        self,
        feature: Optional[Union[List[FeatureNameType], FeatureNameType]] = None,
        resolver: Optional[Union[List[ResolverNameType], ResolverNameType]] = None,
        feature_tag: Optional[Union[List[str], str]] = None,
        resolver_tag: Optional[Union[List[str], str]] = None,
        operation_id: Optional[Union[List[str], str]] = None,
        query_name: Optional[Union[List[str], str]] = None,
        feature_status: Optional[Literal["success", "failure"]] = None,
        resolver_status: Optional[Literal["success", "failure"]] = None,
        cron_status: Optional[Literal["success", "failure"]] = None,
        migration_status: Optional[Literal["success", "failure"]] = None,
        query_status: Optional[Literal["success", "failure"]] = None,
        resolver_type: Optional[Union[List[ResolverType], ResolverType]] = None,
        cache_hit: Optional[bool] = None,
        is_null: Optional[bool] = None,
    ) -> "Series":
        return self._where(
            feature_name=feature,
            resolver_name=resolver,
            feature_tag=feature_tag,
            resolver_tag=resolver_tag,
            operation_id=operation_id,
            query_name=query_name,
            feature_status=feature_status,
            resolver_status=resolver_status,
            cron_status=cron_status,
            migration_status=migration_status,
            query_status=query_status,
            resolver_type=resolver_type,
            cache_hit=cache_hit,
            is_null=is_null,
            equals=True,
        )

    def where_not(
        self,
        feature: Optional[Union[List[FeatureNameType], Union[FeatureWrapper, str]]] = None,
        resolver: Optional[Union[List[ResolverNameType], ResolverNameType]] = None,
        feature_tag: Optional[Union[List[str], str]] = None,
        resolver_tag: Optional[Union[List[str], str]] = None,
        operation_id: Optional[Union[List[str], str]] = None,
        query_name: Optional[Union[List[str], str]] = None,
        feature_status: Optional[Literal["success", "failure"]] = None,
        resolver_status: Optional[Literal["success", "failure"]] = None,
        cron_status: Optional[Literal["success", "failure"]] = None,
        migration_status: Optional[Literal["success", "failure"]] = None,
        query_status: Optional[Literal["success", "failure"]] = None,
        resolver_type: Optional[Union[List[ResolverType], ResolverType]] = None,
        cache_hit: Optional[bool] = None,
        is_null: Optional[bool] = None,
    ) -> "Series":
        return self._where(
            feature_name=feature,
            resolver_name=resolver,
            feature_tag=feature_tag,
            resolver_tag=resolver_tag,
            operation_id=operation_id,
            query_name=query_name,
            feature_status=feature_status,
            resolver_status=resolver_status,
            cron_status=cron_status,
            migration_status=migration_status,
            query_status=query_status,
            resolver_type=resolver_type,
            cache_hit=cache_hit,
            is_null=is_null,
            equals=False,
        )

    def _where(
        self,
        feature_name: Optional[Union[List[FeatureNameType], Union[FeatureWrapper, str]]] = None,
        resolver_name: Optional[Union[List[ResolverNameType], ResolverNameType]] = None,
        feature_tag: Optional[Union[List[str], str]] = None,
        resolver_tag: Optional[Union[List[str], str]] = None,
        operation_id: Optional[Union[List[str], str]] = None,
        query_name: Optional[Union[List[str], str]] = None,
        feature_status: Optional[Literal["success", "failure"]] = None,
        resolver_status: Optional[Literal["success", "failure"]] = None,
        cron_status: Optional[Literal["success", "failure"]] = None,
        migration_status: Optional[Literal["success", "failure"]] = None,
        query_status: Optional[Literal["success", "failure"]] = None,
        resolver_type: Optional[Union[List[ResolverType], ResolverType]] = None,
        cache_hit: Optional[bool] = None,
        is_null: Optional[bool] = None,
        equals: Optional[bool] = True,
    ) -> "Series":
        copy = self._copy_with()
        success_dict = {"success": True, "failure": False}
        if feature_name:
            feature_name = [feature_name] if not isinstance(feature_name, list) else feature_name
            copy = copy._feature_name_filter(*feature_name, equals=equals)
        if resolver_name:
            resolver_name = [resolver_name] if not isinstance(resolver_name, list) else resolver_name
            copy = copy._resolver_name_filter(*resolver_name, equals=equals)
        if feature_tag:
            feature_tag = [feature_tag] if isinstance(feature_tag, str) else feature_tag
            copy = copy._string_filter(*feature_tag, kind="feature_tag", equals=equals)
        if resolver_tag:
            resolver_tag = [resolver_tag] if isinstance(resolver_tag, str) else resolver_tag
            copy = copy._string_filter(*resolver_tag, kind="resolver_tag", equals=equals)
        if operation_id:
            operation_id = [operation_id] if isinstance(operation_id, str) else operation_id
            copy = copy._string_filter(*operation_id, kind="operation_id", equals=equals)
        if query_name:
            query_name = [query_name] if isinstance(query_name, str) else query_name
            copy = copy._string_filter(*query_name, kind="query_name", equals=equals)
        if feature_status:
            copy = copy._status_filter(kind="feature", success=success_dict[feature_status] == equals)
        if resolver_status:
            copy = copy._status_filter(kind="resolver", success=success_dict[resolver_status] == equals)
        if cron_status:
            copy = copy._status_filter(kind="cron", success=success_dict[cron_status] == equals)
        if migration_status:
            copy = copy._status_filter(kind="migration", success=success_dict[migration_status] == equals)
        if query_status:
            copy = copy._status_filter(kind="query", success=success_dict[query_status] == equals)
        if resolver_type:
            resolver_type = [resolver_type] if not isinstance(resolver_type, list) else resolver_type
            copy = copy.with_resolver_type_filter(*resolver_type, equals=equals)
        if cache_hit is not None:
            copy = copy._true_false_filter(kind="cache_hit", value=equals)
        if is_null is not None:
            copy = copy._true_false_filter(kind="null", value=equals)
        return copy

    def with_feature_name_filter(self, *features: Tuple[FeatureNameType]) -> "Series":
        return self._feature_name_filter(*features, equals=True)

    def without_feature_name_filter(self, *features: Tuple[FeatureNameType]) -> "Series":
        return self._feature_name_filter(*features, equals=False)

    def _feature_name_filter(self, *features: Tuple[FeatureNameType], equals: bool) -> "Series":
        if not features:
            raise ValueError(f"One or more Chalk Features must be supplied.")
        copy = self._copy_with()
        comparator = ComparatorKind.EQ if equals else ComparatorKind.NEQ
        if len(features) == 1 or not equals:
            for feature in features:
                value = str(feature) if isinstance(feature, FeatureWrapper) else feature
                metric_filter = MetricFilter(kind=FilterKind.FEATURE_NAME, comparator=comparator, value=[value])
                copy._filters.append(metric_filter)
            if len(features) == 1:
                feature = features[0]
                copy._entity_id = value = str(feature) if isinstance(feature, FeatureWrapper) else feature
                copy._entity_kind = ChartLinkKind.feature
        else:
            value = [str(feature) if isinstance(feature, FeatureWrapper) else feature for feature in features]
            metric_filter = MetricFilter(kind=FilterKind.FEATURE_NAME, comparator=ComparatorKind.ONE_OF, value=value)
            copy._filters.append(metric_filter)
        return copy

    def with_resolver_name_filter(self, *resolvers: Tuple[ResolverNameType]) -> "Series":
        return self._resolver_name_filter(*resolvers, equals=True)

    def without_resolver_name_filter(self, *resolvers: Tuple[ResolverNameType]) -> "Series":
        return self._resolver_name_filter(*resolvers, equals=False)

    def _resolver_name_filter(self, *resolvers: Tuple[ResolverNameType], equals: bool) -> "Series":
        if not resolvers:
            raise ValueError(f"One or more Chalk Resolvers must be supplied.")
        copy = self._copy_with()
        comparator = ComparatorKind.EQ if equals else ComparatorKind.NEQ
        if len(resolvers) == 1 or not equals:
            for resolver in resolvers:
                value = resolver.fqn if isinstance(resolver, Resolver) else resolver
                metric_filter = MetricFilter(kind=FilterKind.RESOLVER_NAME, comparator=comparator, value=[value])
                copy._filters.append(metric_filter)
            if len(resolvers) == 1:
                resolver = resolvers[0]
                copy._entity_id = resolver.fqn if isinstance(resolver, Resolver) else resolver
                copy._entity_kind = ChartLinkKind.resolver
        else:
            value = [resolver.fqn if isinstance(resolver, Resolver) else resolver for resolver in resolvers]
            metric_filter = MetricFilter(kind=FilterKind.RESOLVER_NAME, comparator=ComparatorKind.ONE_OF, value=value)
            copy._filters.append(metric_filter)
        return copy

    def with_feature_tag_filter(self, *tags: Tuple[str]) -> "Series":
        return self._string_filter(*tags, kind="feature_tag", equals=True)

    def without_feature_tag_filter(self, *tags: Tuple[str]) -> "Series":
        return self._string_filter(*tags, kind="feature_tag", equals=False)

    def with_resolver_tag_filter(self, *tags: Tuple[str]) -> "Series":
        return self._string_filter(*tags, kind="resolver_tag", equals=True)

    def without_resolver_tag_filter(self, *tags: Tuple[str]) -> "Series":
        return self._string_filter(*tags, kind="resolver_tag", equals=False)

    def with_operation_id_filter(self, *operation_ids: Tuple[str]) -> "Series":
        return self._string_filter(*operation_ids, kind="operation_id", equals=True)

    def without_operation_id_filter(self, *operation_ids: Tuple[str]) -> "Series":
        return self._string_filter(*operation_ids, kind="operation_id", equals=False)

    def with_query_name_filter(self, *query_names: Tuple[str]) -> "Series":
        return self._string_filter(*query_names, kind="query_name", equals=True)

    def without_query_name_filter(self, *query_names: Tuple[str]) -> "Series":
        return self._string_filter(*query_names, kind="query_name", equals=False)

    def _string_filter(self, *strings: Tuple[str], kind: str, equals=True):
        if not strings:
            raise ValueError(f"One or more arguments must be supplied for this filter")
        copy = self._copy_with()
        filter_kind = self._get_filter_string_type(kind)
        if not filter_kind:
            raise ValueError(f"no filter for '{kind}' found")
        comparator = ComparatorKind.EQ if equals else ComparatorKind.NEQ
        if len(strings) == 1 or not equals:
            for string in strings:
                metric_filter = MetricFilter(kind=filter_kind, comparator=comparator, value=[string])
                copy._filters.append(metric_filter)
            if len(strings) == 1 and kind == "query_name":
                copy._entity_id = strings[0]
                copy._entity_kind = ChartLinkKind.query
        else:
            metric_filter = MetricFilter(kind=filter_kind, comparator=ComparatorKind.ONE_OF, value=list(strings))
            copy._filters.append(metric_filter)
        return copy

    def with_resolver_type_filter(
        self, *resolver_types: Tuple[Literal["online", "offline", "stream"]], equals=True
    ) -> "Series":
        if not resolver_types:
            raise ValueError(f"One or more resolver types from 'online', 'offline', or 'stream' must be supplied")
        if not set(resolver_types).issubset(["online", "offline", "stream"]):
            raise ValueError(f"Resolver types '{resolver_types}' must be one of 'online', 'offline', or 'stream'")
        copy = self._copy_with()
        comparator = ComparatorKind.EQ if equals else ComparatorKind.NEQ
        if len(resolver_types) == 1 or not equals:
            for resolver_type in resolver_types:
                metric_filter = MetricFilter(
                    kind=FilterKind.ONLINE_OFFLINE, comparator=comparator, value=[resolver_type]
                )
                copy._filters.append(metric_filter)
        else:
            metric_filter = MetricFilter(
                kind=FilterKind.ONLINE_OFFLINE, comparator=ComparatorKind.ONE_OF, value=list(resolver_types)
            )
            copy._filters.append(metric_filter)
        return copy

    def with_null_filter(self) -> "Series":
        return self._true_false_filter(kind="null", value=True)

    def without_null_filter(self) -> "Series":
        return self._true_false_filter(kind="null", value=False)

    def with_cache_hit_filter(self) -> "Series":
        return self._true_false_filter(kind="cache_hit", value=True)

    def without_cache_hit_filter(self) -> "Series":
        return self._true_false_filter(kind="cache_hit", value=False)

    def _true_false_filter(self, kind: str, value: bool) -> "Series":
        copy = self._copy_with()
        filter_kind = FilterKind.IS_NULL if kind == "null" else FilterKind.CACHE_HIT
        value = "true" if value else "false"
        metric_filter = MetricFilter(kind=filter_kind, comparator=ComparatorKind.EQ, value=[value])
        copy._filters.append(metric_filter)
        return copy

    def with_feature_status_filter(self, success=True) -> "Series":
        return self._status_filter(kind="feature", success=success)

    def with_resolver_status_filter(self, success=True) -> "Series":
        return self._status_filter(kind="resolver", success=success)

    def with_cron_status_filter(self, success=True) -> "Series":
        return self._status_filter(kind="cron", success=success)

    def with_query_status_filter(self, success=True) -> "Series":
        return self._status_filter(kind="query", success=success)

    def _status_filter(self, kind: str, success: bool) -> "Series":
        copy = self._copy_with()
        value = "success" if success else "failure"
        filter_kind = self._get_filter_status_type(kind)
        if not filter_kind:
            raise ValueError(f"no status filter for '{kind}' found")
        metric_filter = MetricFilter(kind=filter_kind, comparator=ComparatorKind.EQ, value=[value])
        copy._filters.append(metric_filter)
        return copy

    def with_filter(
        self, kind: Union[FilterKind, str], comparator: Union[ComparatorKind, str], value: Union[List[str], str]
    ) -> "Series":
        copy = self._copy_with()
        kind = FilterKind(kind.upper())
        comparator = ComparatorKind(comparator.upper())
        value = [value] if isinstance(value, str) else value
        metric_filter = MetricFilter(kind=kind, comparator=comparator, value=value)
        copy._filters.append(metric_filter)
        return copy

    def with_window_function(self, window_function: Union[WindowFunctionKind, str]) -> "Series":

        copy = self._copy_with()
        copy._window_function = WindowFunctionKind(window_function.upper())
        return copy

    def with_group_by(self, group_by: Union[GroupByKind, str]) -> "Series":
        copy = self._copy_with()
        group_by = GroupByKind(group_by.upper())
        copy._group_by.append(group_by)
        return copy

    def with_group_by_feature_status(self) -> "Series":
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.FEATURE_STATUS)
        return copy

    def with_group_by_feature_name(self) -> "Series":
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.FEATURE_NAME)
        return copy

    def with_group_by_is_null(self) -> "Series":
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.IS_NULL)
        return copy

    def with_group_by_resolver_status(self) -> "Series":
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.RESOLVER_STATUS)
        return copy

    def with_group_by_resolver_name(self) -> "Series":
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.RESOLVER_NAME)
        return copy

    def with_group_by_query_status(self) -> "Series":
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.QUERY_STATUS)
        return copy

    def with_group_by_query_name(self) -> "Series":
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.QUERY_NAME)
        return copy

    def with_group_by_resolver_type(self) -> "Series":
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.ONLINE_OFFLINE)
        return copy

    def with_group_by_cache_hit(self) -> "Series":
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.CACHE_HIT)
        return copy

    def with_time_shift(self, time_shift: int) -> "Series":
        copy = self._copy_with()
        copy._time_shift = time_shift
        return copy

    def _copy_with(self) -> "Series":
        self_copy = deepcopy(self)
        return self_copy

    def __gt__(self, other) -> ThresholdFunction:
        return ThresholdFunction(self, ">", other)

    def __lt__(self, other) -> ThresholdFunction:
        return ThresholdFunction(self, "<", other)

    def __str__(self) -> str:
        return f"Series(name='{self._name}')"

    def __hash__(self) -> int:
        name = self._name if self._name else "."
        metric = str(self._metric) if self._metric else "."
        filter_strings = (
            sorted([f"{f.kind}.{f.comparator}.{'.'.join(f.value)}" for f in self._filters]) if self._filters else "."
        )
        window_function = str(self._window_function) if self._window_function else "."
        group_by = sorted([str(group_by) for group_by in self._group_by]) if self._group_by else "."
        time_shift = str(self._time_shift) if self._time_shift else "."

        series_string = (
            f"series.{name}.{metric}.{'.'.join(filter_strings)}.{window_function}.{'.'.join(group_by)}.{time_shift}"
        )

        return zlib.crc32(series_string.encode())


class Formula:
    def __init__(
        self,
        name: Optional[str] = None,
        kind: Optional[Union[MetricFormulaKind, str]] = None,
        operands: Optional[Union[SingleSeriesOperand, MultiSeriesOperand, DatasetFeatureOperand]] = None,
    ):
        self._name = name
        self._kind = MetricFormulaKind(kind.upper()) if kind else None
        self._operands = operands

    def with_name(self, name: str) -> "Formula":
        copy = self._copy_with()
        copy._name = name
        return copy

    def with_kind(self, kind: Union[MetricFormulaKind, str]) -> "Formula":
        copy = self._copy_with()
        copy._kind = MetricFormulaKind(kind.upper())
        return copy

    def with_operands(
        self, operands: Union[SingleSeriesOperand, MultiSeriesOperand, DatasetFeatureOperand]
    ) -> "Formula":
        copy = self._copy_with()
        copy._operands = operands
        return copy

    def _copy_with(self) -> "Formula":
        self_copy = deepcopy(self)
        return self_copy

    def __hash__(self) -> int:
        name = self._name if self._name else "."
        kind = str(self._kind) if self._kind else "."
        operands = ""
        if isinstance(self._operands, SingleSeriesOperand):
            operands = self._operands.operand
        elif isinstance(self._operands, MultiSeriesOperand):
            operands = self._operands.operands
        elif isinstance(self._operands, DatasetFeatureOperand):
            operands = f"{self._operands.dataset}.{self._operands.feature}"

        formula_string = f"formula.{name}.{kind}.{operands}"

        return zlib.crc32(formula_string.encode())


class Trigger:
    def __init__(
        self,
        name: str,
        severity: Optional[Union[AlertSeverityKind, str]] = None,
        threshold_position: Optional[Union[ThresholdPosition, str]] = None,
        threshold_value: Optional[float] = None,
        series_name: Optional[str] = None,
        channel_name: Optional[str] = None,
    ):
        self._name = name
        self._severity = severity
        self._threshold_position = threshold_position
        self._threshold_value = threshold_value
        self._series_name = series_name
        self._channel_name = channel_name

    def with_name(self, name: str) -> "Trigger":
        copy = self._copy_with()
        copy._name = name
        return copy

    def with_severity(self, severity: Union[AlertSeverityKind, str]) -> "Trigger":
        copy = self._copy_with()
        copy._severity = AlertSeverityKind(severity.upper())
        return copy

    def with_critical_severity(self) -> "Trigger":
        copy = self._copy_with()
        copy._severity = AlertSeverityKind.CRITICAL
        return copy

    def with_error_severity(self) -> "Trigger":
        copy = self._copy_with()
        copy._severity = AlertSeverityKind.ERROR
        return copy

    def with_warning_severity(self) -> "Trigger":
        copy = self._copy_with()
        copy._severity = AlertSeverityKind.WARNING
        return copy

    def with_info_severity(self) -> "Trigger":
        copy = self._copy_with()
        copy._severity = AlertSeverityKind.INFO
        return copy

    def with_threshold_position(self, threshold_position: Union[ThresholdPosition, str]) -> "Trigger":
        copy = self._copy_with()
        copy._threshold_position = ThresholdPosition(threshold_position.upper())
        return copy

    def with_threshold_value(self, threshold_value: float) -> "Trigger":
        copy = self._copy_with()
        copy._threshold_value = threshold_value
        return copy

    def with_series_name(self, series_name: str) -> "Trigger":
        copy = self._copy_with()
        copy._series_name = series_name
        return copy

    def with_channel_name(self, channel_name: str) -> "Trigger":
        copy = self._copy_with()
        copy._channel_name = channel_name
        return copy

    def _copy_with(self) -> "Trigger":
        self_copy = deepcopy(self)
        return self_copy

    def __str__(self) -> str:
        return f"Trigger(name='{self._name}')"

    def __hash__(self) -> int:
        name = self._name if self._name else "."
        severity = str(self._severity) if self._severity else "."
        threshold_position = str(self._threshold_position) if self._threshold_position else "."
        threshold_value = str(self._threshold_value) if self._threshold_value else "."
        series_name = self._series_name if self._series_name else "."
        channel_name = self._channel_name if self._channel_name else "."

        trigger_string = (
            f"trigger.{name}.{severity}.{threshold_position}." f"{threshold_value}.{series_name}.{channel_name}"
        )

        return zlib.crc32(trigger_string.encode())


def _copy_with(function: Callable):
    def inner(self, *args, **kwargs):
        copy = deepcopy(self)
        if not self._keep:
            if self in Chart._registry:
                Chart._registry.remove(self)
        return_copy = function(copy, *args, **kwargs)
        Chart._registry.add(return_copy)
        return return_copy

    return inner


# MetricConfigGQL
class Chart:
    _registry: ClassVar[Set[Union[str, "Chart"]]] = set()

    def __init__(self, name: str, window_period: Optional[str] = None, keep: Optional[bool] = False):
        self._name = name
        self._window_period = window_period
        self._series: List[Series] = []
        self._formulas: List[Formula] = []
        self._trigger = None
        self._keep = keep
        self._entity_id = None
        self._entity_kind = ChartLinkKind.manual
        Chart._registry.add(self)

    @_copy_with
    def with_window_period(self, window_period: str) -> "Chart":
        parse_chalk_duration(window_period)
        self._window_period = window_period
        return self

    def with_feature_request_count_series(
        self, series_name: str, window_function: Union[WindowFunctionKind, str], time_shift: Optional[int] = None
    ) -> "Chart":
        return self.with_series(
            name=series_name,
            metric=MetricKind.FEATURE_REQUEST_COUNT,
            window_function=window_function,
            time_shift=time_shift,
        )

    def with_feature_latency_series(
        self, series_name: str, window_function: Union[WindowFunctionKind, str], time_shift: Optional[int] = None
    ) -> "Chart":
        return self.with_series(
            name=series_name, metric=MetricKind.FEATURE_LATENCY, window_function=window_function, time_shift=time_shift
        )

    def with_feature_staleness_series(
        self, series_name: str, window_function: Union[WindowFunctionKind, str], time_shift: Optional[int] = None
    ) -> "Chart":
        return self.with_series(
            name=series_name,
            metric=MetricKind.FEATURE_STALENESS,
            window_function=window_function,
            time_shift=time_shift,
        )

    def with_feature_value_series(
        self, series_name: str, window_function: Union[WindowFunctionKind, str], time_shift: Optional[int] = None
    ) -> "Chart":
        return self.with_series(
            name=series_name, metric=MetricKind.FEATURE_VALUE, window_function=window_function, time_shift=time_shift
        )

    def with_feature_write_series(
        self, series_name: str, window_function: Union[WindowFunctionKind, str], time_shift: Optional[int] = None
    ) -> "Chart":
        return self.with_series(
            name=series_name, metric=MetricKind.FEATURE_WRITE, window_function=window_function, time_shift=time_shift
        )

    def with_feature_null_ratio_series(self, series_name: str, time_shift: Optional[int] = None) -> "Chart":
        return self.with_series(
            name=series_name,
            metric=MetricKind.FEATURE_NULL_RATIO,
            time_shift=time_shift,
        )

    def with_resolver_request_count_series(
        self, series_name: str, window_function: Union[WindowFunctionKind, str], time_shift: Optional[int] = None
    ) -> "Chart":
        return self.with_series(
            name=series_name,
            metric=MetricKind.RESOLVER_REQUEST_COUNT,
            window_function=window_function,
            time_shift=time_shift,
        )

    def with_resolver_latency_series(
        self, series_name: str, window_function: Union[WindowFunctionKind, str], time_shift: Optional[int] = None
    ) -> "Chart":
        return self.with_series(
            name=series_name, metric=MetricKind.RESOLVER_LATENCY, window_function=window_function, time_shift=time_shift
        )

    def with_resolver_success_ratio_series(self, series_name: str, time_shift: Optional[int] = None) -> "Chart":
        return self.with_series(
            name=series_name,
            metric=MetricKind.RESOLVER_SUCCESS_RATIO,
            time_shift=time_shift,
        )

    def with_query_count_series(
        self, series_name: str, window_function: Union[WindowFunctionKind, str], time_shift: Optional[int] = None
    ) -> "Chart":
        return self.with_series(
            name=series_name, metric=MetricKind.QUERY_COUNT, window_function=window_function, time_shift=time_shift
        )

    def with_query_latency_series(
        self, series_name: str, window_function: Union[WindowFunctionKind, str], time_shift: Optional[int] = None
    ) -> "Chart":
        return self.with_series(
            name=series_name, metric=MetricKind.QUERY_LATENCY, window_function=window_function, time_shift=time_shift
        )

    def with_query_success_ratio_series(self, series_name: str, time_shift: Optional[int] = None) -> "Chart":
        return self.with_series(
            name=series_name,
            metric=MetricKind.QUERY_SUCCESS_RATIO,
            time_shift=time_shift,
        )

    def with_billing_inference_series(self, series_name: str, time_shift: Optional[int] = None) -> "Chart":
        return self.with_series(
            name=series_name,
            metric=MetricKind.BILLING_INFERENCE,
            time_shift=time_shift,
        )

    def with_billing_cron_series(self, series_name: str, time_shift: Optional[int] = None) -> "Chart":
        return self.with_series(name=series_name, metric=MetricKind.BILLING_CRON, time_shift=time_shift)

    def with_billing_migration_series(self, series_name: str, time_shift: Optional[int] = None) -> "Chart":
        return self.with_series(
            name=series_name,
            metric=MetricKind.BILLING_MIGRATION,
            time_shift=time_shift,
        )

    def with_cron_count_series(self, series_name: str, time_shift: Optional[int] = None) -> "Chart":
        return self.with_series(name=series_name, metric=MetricKind.CRON_COUNT, time_shift=time_shift)

    def with_cron_latency_series(
        self, series_name: str, window_function: Union[WindowFunctionKind, str], time_shift: Optional[int] = None
    ) -> "Chart":
        return self.with_series(
            name=series_name, metric=MetricKind.CRON_LATENCY, window_function=window_function, time_shift=time_shift
        )

    def with_stream_messages_processed_series(
        self, series_name: str, window_function: Union[WindowFunctionKind, str], time_shift: Optional[int] = None
    ) -> "Chart":
        return self.with_series(
            name=series_name,
            metric=MetricKind.STREAM_MESSAGES_PROCESSED,
            window_function=window_function,
            time_shift=time_shift,
        )

    def with_stream_message_latency_series(
        self, series_name: str, window_function: Union[WindowFunctionKind, str], time_shift: Optional[int] = None
    ) -> "Chart":
        return self.with_series(
            name=series_name,
            metric=MetricKind.STREAM_MESSAGE_LATENCY,
            window_function=window_function,
            time_shift=time_shift,
        )

    def with_stream_windows_processed_series(
        self, series_name: str, window_function: Union[WindowFunctionKind, str], time_shift: Optional[int] = None
    ) -> "Chart":
        return self.with_series(
            name=series_name,
            metric=MetricKind.STREAM_WINDOWS_PROCESSED,
            window_function=window_function,
            time_shift=time_shift,
        )

    def with_stream_window_latency_series(
        self, series_name: str, window_function: Union[WindowFunctionKind, str], time_shift: Optional[int] = None
    ) -> "Chart":
        return self.with_series(
            name=series_name,
            metric=MetricKind.STREAM_WINDOW_LATENCY,
            window_function=window_function,
            time_shift=time_shift,
        )

    @overload
    def with_series(self, /, series: Series, **kwargs) -> "Chart":
        ...

    @overload
    def with_series(self, /, name: str, **kwargs) -> "Chart":
        ...

    @_copy_with
    def with_series(self, /, name: Optional[str] = None, series: Optional[Series] = None, **kwargs) -> "Chart":
        if series:
            if not isinstance(series, Series):
                raise ValueError(f"'series' value '{series}' must be a Series object")
            self._series.append(series)
            if series._entity_id:
                self._entity_id = series._entity_id
                self._entity_kind = series._entity_kind
            return self
        if name:
            if not isinstance(name, str):
                raise ValueError(f"'name' value '{name}' must be a string")
            new_series = Series(name=name, **kwargs)
            self._series.append(new_series)
            return self
        raise ValueError("Either a 'name' for a new series or an existing Series 'series' must be supplied")

    def get_series(self, series_name: str) -> Series:
        for series in self._series:
            if series._name == series_name:
                return series
        raise ValueError(f"No series named '{series_name}' exists in Chart '{self._name}'")

    def _get_series_index(self, series_name: str) -> Tuple[int, Series]:
        for i, series in enumerate(self._series):
            if series._name == series_name:
                return i, series
        raise ValueError(f"No series named '{series_name}' exists in Chart '{self._name}'")

    @overload
    def with_formula(self, /, formula: Formula, **kwargs) -> "Chart":
        ...

    @overload
    def with_formula(self, /, name: str, **kwargs) -> "Chart":
        ...

    @_copy_with
    def with_formula(self, /, name: Optional[str] = None, formula: Optional[Formula] = None, **kwargs) -> "Chart":
        if formula:
            if not isinstance(formula, Formula):
                raise ValueError(f"'formula' value '{formula}' must be a Formula object")
            self._formulas.append(formula)
            return self
        if name:
            if not isinstance(name, str):
                raise ValueError(f"'name' value '{name}' must be a string")
            new_formula = Formula(**kwargs)
            self._formulas.append(new_formula)
            return self
        raise ValueError("Either a 'name' for a new formula or an existing Formula 'formula' must be supplied")

    @_copy_with
    def with_trigger(
        self,
        expression: ThresholdFunction,
        trigger_name: str,
        severity: Optional[Union[AlertSeverityKind, str]] = None,
        channel_name: Optional[str] = None,
    ) -> "Chart":
        if not isinstance(expression.lhs, Series):
            raise ValueError(f"Left hand side of expression '{expression.lhs}' must be a Series")
        threshold_position = ThresholdPosition.ABOVE if expression.operation == ">" else ThresholdPosition.BELOW
        trigger = Trigger(
            name=trigger_name,
            severity=AlertSeverityKind(severity.upper()),
            threshold_position=threshold_position,
            threshold_value=expression.rhs,
            series_name=expression.lhs._name,
            channel_name=channel_name,
        )
        self._trigger = trigger
        return self

    @_copy_with
    def with_feature_name_filter(
        self,
        *features: Tuple[FeatureNameType],
        series_name: str,
    ) -> "Chart":
        series_index, series = self._get_series_index(series_name=series_name)
        self._series[series_index] = series.with_feature_name_filter(*features)
        return self

    @_copy_with
    def without_feature_name_filter(
        self,
        *features: Tuple[FeatureNameType],
        series_name: str,
    ) -> "Chart":
        series_index, series = self._get_series_index(series_name=series_name)
        self._series[series_index] = series.without_feature_name_filter(*features)
        return self

    @_copy_with
    def with_resolver_name_filter(
        self,
        *resolvers: Tuple[ResolverNameType],
        series_name: str,
    ) -> "Chart":
        series_index, series = self._get_series_index(series_name=series_name)
        self._series[series_index] = series.with_resolver_name_filter(*resolvers)
        return self

    @_copy_with
    def without_resolver_name_filter(
        self,
        *resolvers: Tuple[ResolverNameType],
        series_name: str,
    ) -> "Chart":
        series_index, series = self._get_series_index(series_name=series_name)
        self._series[series_index] = series.without_resolver_name_filter(*resolvers)
        return self

    @_copy_with
    def with_feature_tag_filter(
        self,
        *tags: Tuple[str],
        series_name: str,
    ) -> "Chart":
        series_index, series = self._get_series_index(series_name=series_name)
        self._series[series_index] = series.with_feature_tag_filter(*tags)
        return self

    @_copy_with
    def without_feature_tag_filter(
        self,
        *tags: Tuple[str],
        series_name: str,
    ) -> "Chart":
        series_index, series = self._get_series_index(series_name=series_name)
        self._series[series_index] = series.without_feature_tag_filter(*tags)
        return self

    @_copy_with
    def with_resolver_tag_filter(
        self,
        *tags: Tuple[str],
        series_name: str,
    ) -> "Chart":
        series_index, series = self._get_series_index(series_name=series_name)
        self._series[series_index] = series.with_resolver_tag_filter(*tags)
        return self

    @_copy_with
    def without_resolver_tag_filter(
        self,
        *tags: Tuple[str],
        series_name: str,
    ) -> "Chart":
        series_index, series = self._get_series_index(series_name=series_name)
        self._series[series_index] = series.without_resolver_tag_filter(*tags)
        return self

    @_copy_with
    def with_operation_id_filter(
        self,
        *operation_ids: Tuple[str],
        series_name: str,
    ) -> "Chart":
        series_index, series = self._get_series_index(series_name=series_name)
        self._series[series_index] = series.with_operation_id_filter(*operation_ids)
        return self

    @_copy_with
    def without_operation_id_filter(
        self,
        *operation_ids: Tuple[str],
        series_name: str,
    ) -> "Chart":
        series_index, series = self._get_series_index(series_name=series_name)
        self._series[series_index] = series.without_operation_id_filter(*operation_ids)
        return self

    @_copy_with
    def with_resolver_type_filter(
        self,
        *resolver_types: Tuple[str],
        series_name: str,
    ) -> "Chart":
        series_index, series = self._get_series_index(series_name=series_name)
        self._series[series_index] = series.with_resolver_type_filter(*resolver_types)
        return self

    @_copy_with
    def with_null_filter(
        self,
        series_name: str,
    ) -> "Chart":
        series_index, series = self._get_series_index(series_name=series_name)
        self._series[series_index] = series.with_null_filter()
        return self

    @_copy_with
    def without_null_filter(
        self,
        series_name: str,
    ) -> "Chart":
        series_index, series = self._get_series_index(series_name=series_name)
        self._series[series_index] = series.without_null_filter()
        return self

    @_copy_with
    def with_cache_hit_filter(
        self,
        series_name: str,
    ) -> "Chart":
        series_index, series = self._get_series_index(series_name=series_name)
        self._series[series_index] = series.with_cache_hit_filter()
        return self

    @_copy_with
    def without_cache_hit_filter(
        self,
        series_name: str,
    ) -> "Chart":
        series_index, series = self._get_series_index(series_name=series_name)
        self._series[series_index] = series.without_cache_hit_filter()
        return self

    @_copy_with
    def with_feature_status_filter(
        self,
        series_name: str,
    ) -> "Chart":
        series_index, series = self._get_series_index(series_name=series_name)
        self._series[series_index] = series.with_feature_status_filter()
        return self

    @_copy_with
    def with_resolver_status_filter(
        self,
        series_name: str,
    ) -> "Chart":
        series_index, series = self._get_series_index(series_name=series_name)
        self._series[series_index] = series.with_resolver_status_filter()
        return self

    @_copy_with
    def with_cron_status_filter(
        self,
        series_name: str,
    ) -> "Chart":
        series_index, series = self._get_series_index(series_name=series_name)
        self._series[series_index] = series.with_cron_status_filter()
        return self

    @_copy_with
    def with_query_status_filter(
        self,
        series_name: str,
    ) -> "Chart":
        series_index, series = self._get_series_index(series_name=series_name)
        self._series[series_index] = series.with_query_status_filter()
        return self

    @_copy_with
    def with_filter(
        self,
        kind: Union[FilterKind, str],
        comparator: Union[ComparatorKind, str],
        value: Union[List[str], str],
        series_name: str,
    ) -> "Chart":
        series_index, series = self._get_series_index(series_name=series_name)
        self._series[series_index] = series.with_filter(kind, comparator, value)
        return self

    @_copy_with
    def with_window_function(self, series_name: str, window_function: Union[WindowFunctionKind, str]) -> "Chart":
        series_index, series = self._get_series_index(series_name=series_name)
        self._series[series_index] = series.with_window_function(window_function)
        return self

    @_copy_with
    def with_group_by(self, series_name: str, group_by: Union[GroupByKind, str]) -> "Chart":
        series_index, series = self._get_series_index(series_name=series_name)
        self._series[series_index] = series.with_group_by(group_by)
        return self

    @_copy_with
    def with_group_by_feature_status(
        self,
        series_name: str,
    ) -> "Chart":
        series_index, series = self._get_series_index(series_name=series_name)
        self._series[series_index] = series.with_group_by_feature_status()
        return self

    @_copy_with
    def with_group_by_feature_name(
        self,
        series_name: str,
    ) -> "Chart":
        series_index, series = self._get_series_index(series_name=series_name)
        self._series[series_index] = series.with_group_by_feature_name()
        return self

    @_copy_with
    def with_group_by_is_null(
        self,
        series_name: str,
    ) -> "Chart":
        series_index, series = self._get_series_index(series_name=series_name)
        self._series[series_index] = series.with_group_by_is_null()
        return self

    @_copy_with
    def with_group_by_resolver_status(
        self,
        series_name: str,
    ) -> "Chart":
        series_index, series = self._get_series_index(series_name=series_name)
        self._series[series_index] = series.with_group_by_resolver_status()
        return self

    @_copy_with
    def with_group_by_resolver_name(
        self,
        series_name: str,
    ) -> "Chart":
        series_index, series = self._get_series_index(series_name=series_name)
        self._series[series_index] = series.with_group_by_resolver_name()
        return self

    @_copy_with
    def with_group_by_query_status(
        self,
        series_name: str,
    ) -> "Chart":
        series_index, series = self._get_series_index(series_name=series_name)
        self._series[series_index] = series.with_group_by_query_status()
        return self

    @_copy_with
    def with_group_by_query_name(
        self,
        series_name: str,
    ) -> "Chart":
        series_index, series = self._get_series_index(series_name=series_name)
        self._series[series_index] = series.with_group_by_query_name()
        return self

    @_copy_with
    def with_group_by_resolver_type(
        self,
        series_name: str,
    ) -> "Chart":
        series_index, series = self._get_series_index(series_name=series_name)
        self._series[series_index] = series.with_group_by_resolver_type()
        return self

    @_copy_with
    def with_group_by_cache_hit(
        self,
        series_name: str,
    ) -> "Chart":
        series_index, series = self._get_series_index(series_name=series_name)
        self._series[series_index] = series.with_group_by_cache_hit()
        return self

    @_copy_with
    def with_time_shift(
        self,
        time_shift: int,
        series_name: str,
    ) -> "Chart":
        series_index, series = self._get_series_index(series_name=series_name)
        self._series[series_index] = series.with_time_shift(time_shift)
        return self

    @_copy_with
    def with_feature_link(self, feature: FeatureNameType) -> "Chart":
        self._entity_kind = ChartLinkKind.feature
        self._entity_id = str(feature) if isinstance(feature, FeatureWrapper) else feature
        return self

    @_copy_with
    def with_resolver_link(self, resolver: ResolverNameType) -> "Chart":
        self._entity_kind = ChartLinkKind.resolver
        self._entity_id = resolver.fqn if isinstance(resolver, Resolver) else resolver
        return self

    @_copy_with
    def with_query_link(self, query_name: str) -> "Chart":
        self._entity_kind = ChartLinkKind.query
        self._entity_id = query_name
        return self

    def keep(self) -> "Chart":
        self._keep = True
        return self

    def __str__(self) -> str:
        return f"Chart(name='{self._name}')"

    def __getitem__(self, key: str) -> Union[Series, Formula]:
        for series in self._series:
            if series._name == key:
                return series
        for formula in self._formulas:
            if formula._name == key:
                return formula
        raise ValueError(f"No series or formula named '{key}' exists in Chart {self._name}")

    def __eq__(self, obj):
        return hash(self) == hash(obj)

    def __hash__(self):
        name = self._name if self._name else "."

        window_period = self._window_period if self._window_period else "."
        chart_string = f"chart.{name}.{window_period}"

        series_hash = ".".join(sorted([str(hash(series)) for series in self._series]))
        formulas = ".".join(sorted([str(hash(formula)) for formula in self._formulas]))
        trigger = str(hash(self._trigger)) if self._trigger else "."

        return zlib.crc32(chart_string.encode() + series_hash.encode() + formulas.encode() + trigger.encode())



import re
from datetime import timedelta
from re import Pattern
from typing import Literal, Mapping, Optional, Tuple, Union

Duration = Union[str, timedelta]
"""
Duration is used to describe time periods in
natural langauge. To specify using natural
language, write the count of the unit you would
like, followed by the representation of the unit.

Chalk support the following units:

 | Signifier | Meaning  |
 | --------- | -------- |
 | w         | Weeks    |
 | d         | Days     |
 | h         | Hours    |
 | m         | Minutes  |
 | s         | Seconds  |

As well as the special keyword "infinity"

Examples

 | Signifier   | Meaning                           |
 | ----------- | --------------------------------- |
 | "10h"       | 10 hours                          |
 | "1w 2m"     | 1 week and 2 minutes              |
 | "1h 10m 2s" | 1 hour, 10 minutes, and 2 seconds |
 | "infinity"  | Unbounded time duration           |

Read more at https://docs.chalk.ai/docs/duration
"""

CronTab = str
"""
A schedule defined using the unix-cron
string format (* * * * *).
Values are given in the order below:

| Field        | Values |
| ------------ | ------ |
| Minute       | 0-59   |
| Hour         | 0-23   |
| Day of Month | 1-31   |
| Month        | 1-12   |
| Day of Week  | 0-6    |
"""

ScheduleOptions = Optional[Union[CronTab, Duration, Literal[True]]]


def timedelta_to_duration(t: timedelta) -> Duration:
    s = ""
    if t.days > 0:
        s += f"{t.days}d"
    if t.seconds > 0:
        if len(s) > 0:
            s += " "
        s += f"{t.seconds}s"
    return s


_kwarg_to_regex: Mapping[str, Pattern] = {
    k: re.compile(f"(([0-9]+?){v})")
    for k, v in dict(
        weeks="w",
        days="d",
        hours="h",
        minutes="m",
        seconds="s",
    ).items()
}


def parse_chalk_duration(s: str) -> timedelta:
    def parse(regex: Pattern) -> Tuple[int, str]:
        matched = regex.search(s)
        if matched is None:
            return 0, ""

        return int(matched.groups()[1]), matched.groups()[0]

    parsed_values = {k: parse(unit) for k, unit in _kwarg_to_regex.items()}

    # Check for remaining unparsed input
    remainder = s
    for v in parsed_values.values():
        remainder = remainder.replace(v[1], "", 1)
    remainder = remainder.strip()

    if remainder != "":
        raise ValueError(f"Unparsed portion of duration: '{remainder}' for input: '{s}'")

    return timedelta(**{k: v[0] for k, v in parsed_values.items()})
